#include <naemon/naemon.h>

#include <string.h>

#include "temp_objects.h"

static char *build_command(int sd, struct temp_objects *objs, nagios_macros *mac);
static int run_raw_command(int sd, char *cmd, int timeout, nagios_macros *mac);
static int testthis_qh(int sd, char *buf, unsigned int len);
static int post_config_init(int cb, void *ds);
static void *g_nagios_handle = NULL;

/* Nagios stuff goes below */
NEB_API_VERSION(CURRENT_NEB_API_VERSION);

/**
 * Storage of running job process
 */
struct testthis_command_context_t {
	int sd;
};

int send_string_null_to_fd(int fd, const char *string) {
	ssize_t written;
	ssize_t n;
	ssize_t len;

	written = 0;
	len = strlen(string)+1; /* Include null byte */

	while (written < len) {
		n = (ssize_t)write(fd, string+written, len-written);
		if (n < 0) {
			return n;
		}
		written += n;
	}
	return 0;
}


/**
 * Callback when result is done. Needs to clean up arg.
 */
static void run_callback(wproc_result *wpres, void *arg, int flags) {
	struct testthis_command_context_t *ctx = (struct testthis_command_context_t *)arg;

	if( wpres != NULL ) {
		char *buf = kvvec_to_ekvstr(wpres->response);

		if(send_string_null_to_fd(ctx->sd, buf) < 0) {
			nm_log(NSLOG_RUNTIME_ERROR, "runcmd: Failed to write command response to QH socket: %s\n", strerror(errno));
			nm_free(buf);
			return;
		}
		nm_free(buf);
	} else {
		if (nsock_printf_nul(ctx->sd, "Error: Error running command\n") < 0) {
			/*Failed write() to socket
			 * this assumes the implementation
			 * of nsock_printf_nul uses something
			 * that sets errno.
			 * */
			nm_log(NSLOG_RUNTIME_ERROR, "runcmd: Failed to write to QH socket: %s\n", strerror(errno));
		}
	}
	free(ctx);
}

/**
 * Starts an execution of a processed command line through the worker process.
 */
static int run_raw_command(int sd, char *cmd, int timeout, nagios_macros *mac) {
	struct testthis_command_context_t *ctx = nm_malloc(sizeof(struct testthis_command_context_t));
	int ret = ERROR;
	ctx->sd = sd;
	ret = wproc_run_callback(cmd, timeout, run_callback, (void*)ctx, mac);
	return ret;
}

#define GRAB_MACROS_OR_BUST(FUNC_CALL) do { if(FUNC_CALL == ERROR) { \
	(void)nsock_printf_nul(sd, "Error: Macro grab failed - aborting.\n"); \
	return NULL;} } while (0)
/**
 * Starts an execution of a processed command line through the worker process.
 */
static char *build_command(int sd, struct temp_objects *objs, nagios_macros *mac) {
	char *raw_command = NULL;
	char *processed_command = NULL;
	int macro_options = STRIP_ILLEGAL_MACRO_CHARS | ESCAPE_MACRO_CHARS;

	if( objs->ctc ) GRAB_MACROS_OR_BUST(grab_contact_macros_r(      mac, objs->ctc ));
	if( objs->hst ) GRAB_MACROS_OR_BUST(grab_host_macros_r(         mac, objs->hst ));
	if( objs->hg  ) GRAB_MACROS_OR_BUST(grab_hostgroup_macros_r(    mac, objs->hg  ));
	if( objs->svc ) GRAB_MACROS_OR_BUST(grab_service_macros_r(      mac, objs->svc ));
	if( objs->sg  ) GRAB_MACROS_OR_BUST(grab_servicegroup_macros_r( mac, objs->sg  ));

	/* Fetch command from command object */
	if(( get_raw_command_line_r(mac, objs->cmd, objs->runcommand, &raw_command, macro_options) == ERROR)
			|| raw_command == NULL) {
		(void)nsock_printf_nul(sd, "Error: Raw check command was NULL - aborting.\n");
		return NULL;
	}

	/* process any macros contained in the argument */
	if (( process_macros_r(mac, raw_command, &processed_command, macro_options) == ERROR)
				|| processed_command == NULL) {
		nm_free(raw_command);
		(void)nsock_printf_nul(sd, "Error: Processed check command was NULL - aborting.\n");
		return NULL;
	}
	nm_free(raw_command);
	return processed_command;
}

/**
 * Query handler
 */
static int testthis_qh(int sd, char *buf, unsigned int len) {
    nagios_macros mac;
    char *cmd;
	struct temp_objects *objs = NULL;
    memset(&mac, 0, sizeof(mac));

	if (0 == strcmp(buf, "help")) {
		if ( 0 > nsock_printf_nul(sd,
			"Runs a custom command through nagios worker processes\n"
			"raw <command>   Runs command without any macro expansion\n"
			"run <command>   Runs command with some macros expanded\n"
		)) {
			nm_log(NSLOG_RUNTIME_ERROR, "runcmd: Failed to answer help request on QH socket: %s\n", strerror(errno));
			return QH_CLOSE;
		} else
			return QH_OK;
	}
	if (len > 4 && 0 == memcmp(buf, "raw ", 4)) {
		run_raw_command(sd, buf + 4, 30, &mac);
		return QH_OK;
	}
	if (len > 4 && 0 == memcmp(buf, "run ", 4)) {
		/* Fetch parameters from command line */
		struct kvvec *kvv = ekvstr_to_kvvec(buf+4);
		if(!kvv) {
			nsock_printf_nul(sd, "Error: parsing parameters.\n");
			return QH_OK;
		}
		objs = build_temp_objects( kvv );
		if(!objs) {
			nsock_printf_nul(sd, "Error: building objects.\n");
			kvvec_destroy(kvv, KVVEC_FREE_ALL);
			return QH_OK;
		}
		cmd = build_command(sd, objs, &mac);
		if(!cmd) {
			nsock_printf_nul(sd, "Error: building command.\n");
			clear_volatile_macros_r(&mac);
			destroy_temp_objects(objs);
			kvvec_destroy(kvv, KVVEC_FREE_ALL);
			return QH_OK;
		}
		run_raw_command(sd, cmd, objs->timeout, &mac);
		nm_free( cmd );
		clear_volatile_macros_r(&mac);
		destroy_temp_objects(objs);
		kvvec_destroy(kvv, KVVEC_FREE_ALL);
		return QH_OK;
	}
	if (len > 7 && 0 == memcmp(buf, "expand ", 7)) {
		/* Fetch parameters from command line */
		struct kvvec *kvv = ekvstr_to_kvvec(buf+7);
		if(!kvv) {
			nsock_printf_nul(sd, "Error: parsing parameters.\n");
			return QH_OK;
		}
		objs = build_temp_objects( kvv );
		if(!objs) {
			nsock_printf_nul(sd, "Error: building objects.\n");
			kvvec_destroy(kvv, KVVEC_FREE_ALL);
			return QH_OK;
		}
		cmd = build_command(sd, objs, &mac);
		if(!cmd) {
			nsock_printf_nul(sd, "Error: building command.\n");
			clear_volatile_macros_r(&mac);
			destroy_temp_objects(objs);
			kvvec_destroy(kvv, KVVEC_FREE_ALL);
			return QH_OK;
		}

		if(send_string_null_to_fd(sd, cmd) < 0) {
			nm_log(NSLOG_RUNTIME_ERROR, "runcmd: Failed to write expanded command to QH socket: %s\n", strerror(errno));
		}

		nm_free( cmd );
		clear_volatile_macros_r(&mac);
		destroy_temp_objects(objs);
		kvvec_destroy(kvv, KVVEC_FREE_ALL);
		return QH_OK;
	}
	return QH_INVALID;
}

static int post_config_init(int cb, void *ds) {
	int result;

	if (*(int *)ds != NEBTYPE_PROCESS_EVENTLOOPSTART)
		return 0;

	event_broker_options = BROKER_EVERYTHING;

	neb_deregister_callback(NEBCALLBACK_PROCESS_DATA, post_config_init);

	if ((result = qh_register_handler("runcmd", "Run Command test", 0,
			testthis_qh)) < 0) {
	} else {
	}

	return 0;
}

int nebmodule_init(int flags __attribute__ ((__unused__)), char *args,
		void *handle) {
	g_nagios_handle = handle;

	/* this gets de-registered immediately, so we need to add it manually */
	neb_register_callback(NEBCALLBACK_PROCESS_DATA, g_nagios_handle, 0,
			post_config_init);

	return 0;
}

int nebmodule_deinit(int flags __attribute__ ((__unused__)),
		int reason __attribute__ ((__unused__))) {
	return 0;
}
